 /*
  * UAE - The Un*x Amiga Emulator
  *
  * Virtual File System support - unix version
  *
  * (c) 1997 Gustavo Goedert
  */

#include "sysconfig.h"
#include "sysdeps.h"

#include <ctype.h>

#include "filesys.h"
#include "vfs.h"

typedef struct {
	char *Name;
	ULONG Attributes;
	int Modified;
} VFS_Entry;

int NumberEntrys;
VFS_Entry **Entrys;
char CurrentPath[1024] = "";

int LoadVFS(void);
int LoadPath(char *Directory);
void AddVFS_Entry(VFS_Entry *NewEntry);
VFS_Entry *FindName(char *Name);
void ResetVFS(void);
void SaveVFS(void);

int LoadVFS(void) {
    char TempName[1024];
    FILE *VFSFile;
    int i, Error = 0;
    unsigned short NEntrys, Length;
    VFS_Entry *Entry;
    DIR *dir;
    struct dirent *ent;

    ResetVFS();
    if (access(CurrentPath, D_OK) != 0)
	return(0);
    strcpy(TempName, CurrentPath);
    strcat(TempName, VFSFILENAME);
    if (access(TempName, F_OK) == 0) {
	VFSFile = fopen(TempName, "rb");
	if (VFSFile != NULL) {
	    if (fread(&NEntrys, sizeof(unsigned short), 1, VFSFile) == 1) {
		for (i=0; i<NEntrys; i++) {
		    Entry = malloc(sizeof(VFS_Entry));
		    Entry->Name = NULL;
		    if (Entry == NULL)
			Error = 1;
		    if (!Error)
			if (fread(&Length, sizeof(unsigned short), 1, VFSFile) != 1)
			    Error = 1;
		    if (!Error) {
			Entry->Name = malloc(Length+1);
			if (Entry->Name == NULL)
			    Error = 1;
			if (!Error) {
			    Entry->Name[Length] = '\0';
			    if (fread(Entry->Name, 1, Length, VFSFile) != (unsigned) Length) {
				free(Entry->Name);
				Error = 1;
			    }
			}
		    }
		    if (!Error) {
			if (fread(&(Entry->Attributes), sizeof(ULONG), 1, VFSFile) != 1)
			    Error = 1;
			else
			    Entry->Modified = 1;
		    }
		    if (!Error) {
			strcpy(TempName, CurrentPath);
			strcat(TempName, Entry->Name);
			if (access(TempName, F_OK) == 0)
			    AddVFS_Entry(Entry);
			else {
			    free(Entry->Name);
			    free(Entry);
			}
		    } else {
			if (Entry->Name != NULL)
			    free(Entry->Name);
			free(Entry);
			break;
		    }
		}
		if (Error)
		    ResetVFS();
	    }
	    fclose(VFSFile);
	}
    }
    if ((dir = opendir(CurrentPath)) == NULL)
	return(0);
    while ((ent = readdir(dir)) != NULL) {
	if (FindName(ent->d_name) == NULL) {
	    Entry = malloc(sizeof(VFS_Entry));
	    if (Entry != NULL) {
		Entry->Name = strdup(ent->d_name);
		Entry->Attributes = 0;
		Entry->Modified = 0;
		AddVFS_Entry(Entry);
	    }
	}
    }
    closedir(dir);
    return(1);
}

int LoadPath(char *Directory) {
    char TempName[1024];

    strcpy(TempName, Directory);
    strcat(TempName, "/");
    if (CurrentPath[0]) {
	if (strcasecmp(TempName, CurrentPath) == 0)
	    return(1);
    }
    strcpy(CurrentPath, TempName);
    if (!LoadVFS()) {
	*CurrentPath = '\0';
	return(0);
    }
    return(1);
}

void AddVFS_Entry(VFS_Entry *NewEntry) {
    if (NumberEntrys == 0)
	Entrys = malloc(sizeof(VFS_Entry *));
    else
	Entrys = realloc(Entrys, sizeof(VFS_Entry *)*(NumberEntrys+1));
    if (Entrys != NULL) {
	Entrys[NumberEntrys] = NewEntry;
	NumberEntrys++;
    }
}

VFS_Entry *FindName(char *Name) {
    int i;
    VFS_Entry *Entry;

    for (i = 0; i < NumberEntrys; i++)
	if (strcasecmp(Name, Entrys[i]->Name) == 0)
	    return (Entrys[i]);
    return(NULL);
}

char *CreateRealName(char *Directory, char *VirtualName, int MangleSize) {
    return(VirtualName);
}

char *GetRealName(char *Directory, char *VirtualName) {
    return(VirtualName);
}

char *GetVirtualName(char *Directory, char *RealName) {
    return(RealName);
}

ULONG GetAttributes(char *Directory, char *RealName, mode_t Mode) {
    VFS_Entry *Entry;

    if (LoadPath(Directory)) {
	Entry = FindName(RealName);
	if (Entry != NULL)
	    if (Entry->Modified)
		return(Entry->Attributes);
    }
    return((S_IXUSR & Mode ? 0 : A_FIBF_EXECUTE) |
	   (S_IWUSR & Mode ? 0 : A_FIBF_WRITE) |
#if 0 /* This makes it impossible to overwrite those. */
	   (S_IWUSR & Mode ? 0 : A_FIBF_DELETE) |
#endif
	   (S_IRUSR & Mode ? 0 : A_FIBF_READ));
}

int SetAttributes(char *Directory, char *RealName, mode_t Mode, ULONG Attributes) {
    VFS_Entry *Entry;

    if (LoadPath(Directory)) {
	Entry = FindName(RealName);
	if (Entry != NULL) {
	    Entry->Attributes = Attributes;
	    Entry->Modified = 1;
	    SaveVFS();
	}
    }
    if (Attributes & A_FIBF_READ)
	Mode &= ~S_IRUSR;
    else
	Mode |= S_IRUSR;

    if ((Attributes & A_FIBF_WRITE) != 0 || (Attributes & A_FIBF_DELETE) != 0)
	Mode &= ~S_IWUSR;
    else
	Mode |= S_IWUSR;

    if (Attributes & A_FIBF_EXECUTE)
	Mode &= ~S_IXUSR;
    else
	Mode |= S_IXUSR;
    return(chmod(RealName, Mode));
}


void ResetVFS(void) {
    int i;

    for (i = 0; i < NumberEntrys; i++) {
	free(Entrys[i]->Name);
	free(Entrys[i]);
    }
    if (NumberEntrys)
	free(Entrys);
    NumberEntrys = 0;
}

void CleanDir(char *Directory) {
    strcpy(CurrentPath, Directory);
    strcat(CurrentPath, "/");
    LoadVFS();
    SaveVFS();
    *CurrentPath = '\0';
}

void SaveVFS(void) {
    char TempName[1024];
    int i;
    unsigned short Length = 0;
    FILE *VFSFile = NULL;

    strcpy(TempName, CurrentPath);
    strcat(TempName, VFSFILENAME);
    unlink(TempName);
    for (i = 0; i < NumberEntrys; i++)
	if (Entrys[i]->Modified)
	    Length++;
    if (Length) {
	VFSFile = fopen(TempName, "wb");
	if (VFSFile != NULL) {
	    fwrite(&Length, sizeof(unsigned short), 1, VFSFile);
	    for (i = 0; (i < NumberEntrys) && (VFSFile != NULL); i++) {
		if (Entrys[i]->Modified) {
		    Length = strlen(Entrys[i]->Name);
		    fwrite(&Length, sizeof(unsigned short), 1, VFSFile);
		    fwrite(Entrys[i]->Name, Length, 1, VFSFile);
		    fwrite(&Entrys[i]->Attributes, sizeof(ULONG), 1, VFSFile);
		}
	    }
	}
    }
    if (VFSFile != NULL)
	fclose(VFSFile);
}
